/*                               -*- Mode: C -*-
 * oprs-com.c --
 *
 * Copyright (C) 1994-2012 Francois Felix Ingrand
 *
 * Author          : Francois Felix Ingrand
 * Created On      : Wed Apr 20 16:31:25 1994
 * Last Modified By: Francois Felix Ingrand
 * Last Modified On: Mon Jun 11 16:55:42 2001
 * Update Count    : 352
 * Status          : OK
 *
 *
 */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <time.h>
#include <sys/time.h>


#include <listPack.h>
#include <hashPack.h>


#include "macro.h"

#include "macro-pub.h"

#include "constant-pub.h"
#include "opaque-pub.h"
#include "macro-pub.h"
#include "oprs-type-pub.h"
#include "pu-mk-term_f.h"
#include "pu-parse-tl_f.h"

#include "oprs-type_f-pub.h"
#include "lisp-list_f-pub.h"
#include "fact-goal_f-pub.h"
#include "oprs_f-pub.h"

#include "genom3/c/client.h"

#include "genom-oprs.h"
#include "genom-oprs_f.h"
#include "genom-oprs-com.h"
#include "genom-oprs-com_f.h"

#define COPY_LIST(list) copy_list((list),NOT_NODES)
#define FREE_LIST(list) free_list((list),NOT_NODES)

typedef struct {		/* Request type */
     Symbol name;		/* Name (a symbol) */
     const struct genom_service_info *genom_service_info; /* we have got everything in there...  */
     Encode_Func_Proto encode_func; /* Encode and decode functions */
     Decode_Func_Proto decode_func;
} Rqst_Type;

typedef struct {		/* An Active Request */
     int rqid;			/* its id (mine) */
     Rqst_Type *type;		/* its type */
     Mod *mod;			/* For CSLIB, ist module */
     int ext_rqid;		/* its cslib or tclservclient id */
     PBoolean cslib;            /* are we using cslib or tclservclient */
     PBoolean ir_received;	/* has it received an intermediate reply yet */
} Rqst_Instance;


/* Various hash tables */

#define MOD_HASH_SIZE 16		/* has to be a power of two */
static int mod_hashtable_size = MOD_HASH_SIZE;
static int mod_hashtable_mod = MOD_HASH_SIZE - 1;
HASH mod_hash;

#define RQST_TYPE_HASH_SIZE 32		/* has to be a power of two */
static int rqst_type_hashtable_size = RQST_TYPE_HASH_SIZE;
static int rqst_type_hashtable_mod = RQST_TYPE_HASH_SIZE -1;

static PBoolean oprs_com_verbose = TRUE;


void end_oprs_com ( void );
void init_oprs_com(char *name);



/* Hash functions. */
static int mod_hash_a_string(PString name)
{
     return (hash_a_string(name) & mod_hashtable_mod);
}
static PBoolean mod_match(PString s1, Mod *mod)
{
     return (s1 == mod->name) ;	/* These are atoms, so they can be pointer compared */
}


static int rqst_type_hash_a_string(PString name )
{
     return (hash_a_string(name) & rqst_type_hashtable_mod);
}
static PBoolean rqst_type_match(PString s1, Rqst_Type *rqst_type)
{
     return (s1 == rqst_type->name) ;	/* These are atoms, so they can be pointer compared */
}


/* The global reply message queue or client id. */
//CLIENT_ID rsId;

static void end_com_to_module(OPRS_NODE ignore, Mod *mod)
{
     if (mod->genom_client)		/* if valid genom_client, shutdown it. */
	  mod->genom_client_info->fini(mod->genom_client);
     /* Should we clean up and free everything? */
}

void end_oprs_com()
{
     /* terminate the logging socket server */
     end_server();

     for_all_hashtable(mod_hash, NULL, (SL_PFI)end_com_to_module); /* for all known modules */
}

static Pred_Func_Rec *ir_pred, *fr_pred, *failure_pred;
static Pred_Func_Rec *data_func;

void init_oprs_com(char *name)
{
     /* Init the logging socket server */
     start_server(3140);

     ir_pred = declare_pred_from_symbol(declare_atom("IR"));
     failure_pred = declare_pred_from_symbol(declare_atom("MOD-FAILURE-DATA"));
     fr_pred = declare_pred_from_symbol(declare_atom("FR"));

     data_func = declare_pred_from_symbol(declare_atom("DATA"));

     mod_hash = make_hashtable(mod_hashtable_size, mod_hash_a_string, mod_match);
}

Mod *declare_oprs_module(char * name, const struct genom_client_info *genom_client_info)
{
     Mod *mod = MAKE_OBJECT(Mod);

     mod->name = declare_atom(name); /* the name of this particular instance */
     mod->genom_client = NULL;
     mod->genom_client_info = genom_client_info;
     mod->generic_name = declare_atom((char *)genom_client_info->name); /* the generic name of this module */

     mod->rqst_instance_list = make_list();
     mod->initialized = FALSE;
     mod->poster_hash = make_poster_hashtable();

     mod->rqst_type_hash = make_hashtable(rqst_type_hashtable_size, 
					  rqst_type_hash_a_string, 
					  rqst_type_match);

     if ((get_from_hashtable(mod_hash,mod->name)) != NULL)
	  fprintf(stderr, "genom-oprs:redeclaring the module:%s.\n", mod->name);

     add_to_hashtable(mod_hash, mod, mod->name); /* redeclaring it anyway... */

     return mod;
}

PBoolean init_oprs_mod_com(Mod *mod)
{
     int argc = 2;
     char *argv[3];

     argv[0] = strdup("-i");	/* strdup not needed */
     argv[1] = strdup(mod->name); /* strdup not needed */
     argv[2] = NULL;		/*  this is not really needed a priori */

     if (!mod->initialized) {
	  if ((mod->genom_client = mod->genom_client_info->init(argc,argv)) == NULL)
	       fprintf(stderr, "genom-oprs:init_oprs_mod_com:client_init_func for %s returned NULL.\n", mod->name);
	  else {
	       mod->initialized = TRUE;
//	       mod->name = declare_atom(*(mod->genom_client_info->instance)(mod->genom_client));
	  }
     } else
	  fprintf(stderr, "genom-oprs:init_oprs_mod_com: module: %s, already initialized.\n", mod->name);

     fprintf(stderr, "argv[0] %s , argv[1] %s\n",argv[0], argv[1]);
     
     return mod->initialized;
}

static PBoolean init_oprs_mod_name_com(char *mod_name)
{
     Mod *mod;

     if ((mod = (Mod *)get_from_hashtable(mod_hash,mod_name)) == NULL) {
	  fprintf(stderr, "genom-oprs:init_oprs_mod_com:undeclared module: %s.\n", mod_name);
	  return FALSE;
     } else
	  return init_oprs_mod_com(mod);
}

Term *init_com_to_module(TermList terms)
{
     Term *t1;

     t1 = (Term *)get_list_pos(terms,1);
     if (t1->type != ATOM) {
	  fprintf(stderr,"Expecting an ATOM in init_com_to_module.\n");
	  return build_nil();
     } else {
	  if (init_oprs_mod_name_com(t1->u.id))
	       return build_t();
	  else
	       return build_nil();
     }
}

void init_rqst_type(char *name, Mod *mod, const struct genom_service_info *genom_service_info, 
//genom_request_sendfn genom_request_sendfn, 
		    Encode_Func_Proto encode,// int encode_size,
		    Decode_Func_Proto decode //, int decode_size)
     )
{
     Rqst_Type *rqst_type = MAKE_OBJECT(Rqst_Type);

     rqst_type->name = declare_atom(name);	/* Declare the rqst type as a symbol */
     rqst_type->genom_service_info = genom_service_info;
     rqst_type->encode_func = encode;
     rqst_type->decode_func = decode;

     if ((get_from_hashtable(mod->rqst_type_hash, rqst_type->name)) != NULL)
	  fprintf(stderr, "genom-oprs: WARNING: redeclaring the rqst type:%s for module %s.\n", rqst_type->name, mod->name);

     add_to_hashtable(mod->rqst_type_hash, rqst_type, rqst_type->name);
}

/* Function to post the intermediate reply fact. */
static void post_ir(char *from, Rqst_Instance *rqst_instance, int ir_letter)
{
     TermList terms;

     terms = make_list();
     add_to_tail(terms, build_id(from));
     add_to_tail(terms, build_id(rqst_instance->type->name));
     add_to_tail(terms, build_integer(rqst_instance->rqid));
     add_to_tail(terms, build_integer(ir_letter));
     add_fact(make_fact_from_expr(build_expr_pfr_terms(ir_pred,
						       terms))
	      ,current_oprs);
}

/* Function to post the final reply fact. */

static void post_fr_sreport(char *from, Rqst_Instance *rqst_instance, Term *term, const char *sreport)
{
     TermList terms2;

     terms2 = make_list();
     add_to_tail(terms2, build_id(from));
     add_to_tail(terms2, build_id(rqst_instance->type->name));
     add_to_tail(terms2, build_integer(rqst_instance->rqid));
     add_to_tail(terms2, build_string(sreport));
     add_to_tail(terms2, term);

     add_fact(make_fact_from_expr(build_expr_pfr_terms(fr_pred,
						       terms2)),
	      current_oprs);
}

static void post_fr(char *from, Rqst_Instance *rqst_instance, Term *term, int report)
{
     const char *error = ""; //h2getErrMsg(report, strerr, 64);
     //char *res;

     /* if ((res = strchr(error,'_')) && ((res = strchr(res+1,'_'))) && *(res+1)) */
     /* 	  error = res+1; */
     post_fr_sreport(from, rqst_instance, term, error);
}

void genom_oprs_service_cb(genom_client h, int activity, int error, uint32_t report, void *output, void *cb_data)
{
     Rqst_Instance *rqst_instance = (Rqst_Instance *)cb_data;
     Mod *mod = rqst_instance->mod; 
     Term *term;
     PBoolean del_rqst = FALSE;

     fprintf(stderr, "genom-oprs: genom_service_cb: getting reply  %s, rqid %d, ext_rqid %d, activity %d, error %d, report %d.\n",
	     rqst_instance->type->name, rqst_instance->rqid, rqst_instance->ext_rqid, activity, error, report);
     

     if (error == 0) {
	  if (report == 0) {
	       if (output == NULL){  /* IR */
		    if (!rqst_instance->ir_received) {
			 post_ir(rqst_instance->mod->name, rqst_instance, activity);
			 rqst_instance->ir_received = TRUE;
		    }
	       } else {		/* FR tout est OK */

		    if (rqst_instance->type->genom_service_info->output_size) { /* if size null, nothing to decode */
			 term = (*(rqst_instance->type->decode_func))(output);
// pas a moi de le faire rqst_instance->type->genom_service_info->fini_output(output);

		    } else {
			 term = PUMakeTermNil();
		    }
		    post_fr_sreport(rqst_instance->mod->name, rqst_instance, term, "OK");
		    del_rqst = TRUE;
		    if (oprs_com_verbose) {
			 printf("genom-oprs:  rcv  %s %d\n", rqst_instance->type->name, rqst_instance->rqid);
			 fprintf(stderr,"genom-oprs:  rcv  %s %d\n", rqst_instance->type->name, rqst_instance->rqid);
		    }
		    server_send("[FR %d \"OK\"]", rqst_instance->rqid);
	       }
	  } else {		/* report not OK */
	       const char *error = mod->genom_client_info->strerror(mod->genom_client);
	       term = PUMakeTermNil();
	       
	       post_fr_sreport(rqst_instance->mod->name, rqst_instance, term, error);
	       del_rqst = TRUE;
	  }
     } else {			/* error not OK... cas pathologic... */
	  const char *error = mod->genom_client_info->strerror(mod->genom_client);
	  term = PUMakeTermNil();
	  
	  post_fr_sreport(rqst_instance->mod->name, rqst_instance, term, error);
	  del_rqst = TRUE;
     }

     if (del_rqst) {
	  mod->genom_client_info->clean(h,activity);
	  delete_list_node(rqst_instance->mod->rqst_instance_list,rqst_instance); 
	  /* clean de la requete */
	  FREE_OBJECT(rqst_instance);
     }

     return;
}

static void check_module(OPRS_NODE ignore, Mod *mod)
{
     if (! sl_slist_empty(mod->rqst_instance_list)) { /* if this module has no pending rqst, skip this? */
	  mod->genom_client_info->doevents(mod->genom_client); /* this will call the callback for all new events. */
	  // this could be (and will be) smarter by "selecting" on the
	  // file descriptor from openprs main loop, but it will do for now.
     }
}

PBoolean get_reply_and_check_com(TermList ignore)
{
     Rqst_Instance *rqst_instance;

     static int tick_count = 0;
     struct timeval real_time;
     struct timezone tzp;
     
     if (gettimeofday(&real_time,&tzp) != 0)
	  perror("gettimeofday");
     else
          server_send("[TICK %d %ld%06ld]", tick_count, real_time.tv_sec, real_time.tv_usec);

     ++tick_count;

     for_all_hashtable(mod_hash, NULL, (SL_PFI)check_module); /* for all known modules */

     return FALSE;		/* Never return TRUE, unless we want to stop. */
}

Term *send_request_to_module(TermList terms)
{
     static int rqid = 1;
     Rqst_Type *rqst_type;
     Mod *mod;
     int encode_res, status;
     int ires;
     Term * res;
     char *mod_name;
     char *type_name;

     if (! (PUGetOprsParameters(terms, 2, ATOM, &mod_name, ATOM, &type_name))) {
	  fprintf(stderr, "genom-oprs:send_request_to_module:GetOprsParameters:unable to parse request.\n");
	  res = build_integer(-1);
     } else if ((mod = (Mod *)get_from_hashtable(mod_hash, mod_name)) == NULL) {
	  fprintf(stderr, "genom-oprs:send_request_to_module:could not find module %s.\n", mod_name);
	  res = build_integer(-1);
        } else if ((rqst_type = (Rqst_Type *)get_from_hashtable(mod->rqst_type_hash, type_name)) == NULL) {
	  fprintf(stderr, "genom-oprs:send_request_to_module:could not find request type %s.\n", type_name);
	  res = build_integer(-1);
     } else if (!mod->initialized && !init_oprs_mod_com(mod)) {
	  fprintf(stderr, "genom-oprs:send_request_to_module: module %s not initialized.\n", mod_name);
	  res = build_integer(-1);
     } else {
	  Rqst_Instance *rqst_instance;
	  char *rq_letter;
	  Expression *term;
	  
	  if (rqst_type->genom_service_info->input_size) { /* if size is null, then nothing to encore (even if the function "exists")... */
	       
	       if (! PUGetOprsParametersSpecArg(terms, 3, EXPRESSION, &term)) {
		    fprintf(stderr, "genom-oprs:send_request_to_module:PUGetOprsParametersSpecArg:unable to parse request.\n");
		    return build_integer(-1);
	       }

	       rq_letter = (char *)malloc(rqst_type->genom_service_info->input_size);
	       rqst_type->genom_service_info->init_input(rq_letter);

	       encode_res = (*(rqst_type->encode_func))(
		    term,		/* Term */
		    (void *)(rq_letter)); /* Structure pointer */

	       if (! encode_res) {
		    fprintf(stderr, "genom-oprs:send_request_to_module:problem while encoding data\n\
\tfor request %s for module %s.\n", type_name, mod_name);
		    free(rq_letter);
		    return build_integer(-1);
	       }
	  }

	  rqst_instance = MAKE_OBJECT(Rqst_Instance);
	  rqst_instance->rqid = rqid;
	  rqst_instance->type = rqst_type;
	  rqst_instance->mod = mod;
	  rqst_instance->ir_received = FALSE;

	  ires = rqst_type->genom_service_info->send(mod->genom_client, rq_letter, genom_oprs_service_cb, genom_oprs_service_cb, rqst_instance,  &(rqst_instance->ext_rqid));

	  rqst_type->genom_service_info->fini_input(rq_letter);

	  server_send("[RQ %s %d]", type_name, rqid);
	  if (oprs_com_verbose) {
	       fprintf(stderr,"genom-oprs: *send cslib %s %d\n", type_name, rqid);
	       printf("genom-oprs: *send cslib %s %d\n", type_name, rqid);
	  }
	  add_to_tail(mod->rqst_instance_list, rqst_instance);
	  res = build_integer(rqid);
	  rqid++;
	  }
     return res;
}

Term *set_oprs_com_verbose(TermList terms)
/* This is the action which sends the request. */
{
     Term *t1;

     t1 = (Term *)get_list_pos(terms,1);
     if (t1->type != ATOM) {
	  fprintf(stderr,"Expecting an ATOM in set_oprs_com_verbose.\n");
	  return build_nil();
     } else {
	  oprs_com_verbose = ! (t1->u.id == nil_sym);
	  return build_t();
     }
}

